// S2

// the following lines is supposed(optional) to replace a single line in sketch, the #include"Esp8266MyWiFiUnified.h".
// For the following macros copied to main for easy modify. So, copy these to there and uncomment it could simply done.
#ifndef _ESP8266_MY_WIFI_UNIFIED_DISABLEH_
#define _ESP8266_MY_WIFI_UNIFIED_DISABLEH_
#include <ESP8266WiFi.h>

////////////////// paste the macros here (which is between IoT BEGIN/END .h) //////////////////
// ************************** IoT BEGIN .h ******************************
// ========================== Options and Defs/modify if need ==============================
// -------------------------- for AP or STA ------------------------------
// 1 for ap, 0 for sta.
#define AP_MODE 0
#define STA_MODE (!AP_MODE)

// -------------------------- for ap/sta credential ------------------------------ comment it and put it in .cpp
#if 1
const char *const WIFI_CRED[][2]={
    {"ESP_TESTAP", "87654321"},
    {"Office", "123456"},
    {"Xperia X Performance_bc10", "e848acbcfd98"},
};
#endif // if 1

// -------------------------- for part of hostname ------------------------------
#define LSERVICE_1 "motor"
#define LPSERVICE_1 80

// -------------------------- for default connection names ------------------------------
// i.e., http://hostname.local/
#define HOSTNAME_BASE "esp8266-iot" // for local used.
#define HOSTNAME ((String(HOSTNAME_BASE)+"-"+LSERVICE_1).c_str()) // for local used.
#define REMOTE_HOSTNAME "esp8266-iot-internet" // for remote use.

#define UPDFILE "a.bin" // the remote webpage for updating to local is http://REMOTE_HOSTNAME.local/upd

// -------------------------- other settings ------------------------------
#define HTTPC_GET_DELAY delay(25) // do the delay after an HTTP GET.
#define HTTPC_GET_PROC_DELAY delay(5) // do some delay during an HTTP GET course.
#define HTTPC_CONN_TIMEOUT 3000 // the timeout during a GET, 3000ms.

// all the TMO values is restricted to no more than 35 minutes; see the Timeout() for detail. however,
// cascading could overcome.
#define TMO_IOT_LOOP_PERIOD 75 // IoT polling at each 75ms
#define TMO_AP_C_CONN_PERIOD 30000 // each 30 seconds to show connections by AP
#define TMO_STA_CONN_HEALTH_PERIOD 30000 // each 30 seconds to check the connection health by STA(while ping is alive)
#define TMO_WEB_UPDATE_PERIOD (1000*60*30) // each 30 minutes to find internet update; if cascade twice(no, use count) will become 15hr
#define TMO_MDNS_Q_PERIOD 60000 // each 60 seconds to query services
#define TMO_PING_PERIOD 6000 // each 6 seconds to ping AP when ping failed

#define STA_RECOVER_ALIVE 1 // set 0 if STA connection broken and do not reconnect. 1 to reconnect.

// [mDns] use either of one
//#define USE_MDNS_STANDALONE // no auto hostname
#define USE_MDNS_FULL // auto hostname and query

// [STAtion] use either of one
//#define USE_STA_NATIVE // single AP
#define USE_STA_WIFI_MULTI // multiple APs(o)

// [HTTPClient] use either of one
//#define USE_HTTPC_HTTP_CLIENT
//#define USE_HTTPC_WIFI_CLIENT
#define USE_HTTPC_WIFIHTTP_HYBRID // (o2)
//#define USE_HTTPC_WIFIHTTP_HYBRID_NONBLOCKING // (o1)

// -------------------------- for mdns ------------------------------
typedef struct smDnsRecord{
    const char *service;
    const unsigned port;
} smDnsRecord;

// -------------------------- for mdns services ------------------------------ comment it and put it in .cpp
#if 1
#if defined(USE_MDNS_FULL)
const smDnsRecord mdns_ss[]={ // the services this device provides; note, at least the LSERVICE_1 has already provided.
    {"iot", 80},
    {"https", 443},
    {"test", 81},
    {"motor", 80},
    {"display", 80},
    {"http", 80},
};
#endif // USE_MDNS_FULL

const smDnsRecord mdns_sskn[]={ // the known services this device needs from other devices.
    {"iot", 80},
    {"https", 443},
    {"test", 81},
    {"motor", 80},
    {"display", 80},
    {"http", 80},
};
#endif // if 1


// ************************** IoT END .h ******************************
////////////////// paste the macros here (which is between IoT BEGIN/END .h) //////////////////

#include"Esp8266MyWiFiUnified.h"
#endif // _ESP8266_MY_WIFI_UNIFIED_DISABLEH_


#include"Esp8266MultiPwms.h"
#include"Esp8266Vars2File.h"


//////////////////////////////////////// main function /////////////////////////////////////
// use MultiPwms Lib v.0.9. gpio parking is changed to GPIO3. (however, just set as input could be considered.)
// use "ESP-07S"
// 2021.03.14.

// need to call Start() again if Stop(). setDC is needed too.
// during power-up/reset/reset-after-update has a 6ms all-high of the four pwms. so better to external pull-down.
// adc is used. so remember not to ESP.getVcc() which would damage power-rail.

// configuration/used:
// IDE: NodeMCU0.9(ESP-12), 115200bps, 80MHz, 4MB flash
// pins: TX/RX
// pins: GPIO0/pull-low
// pins: GND/Open
// connector: power +/-
// connector: ADC input 15V max, 15:1.
// power board: 3.3V. max input source 24V
// GPIO5, pull low, to motordrv in1, to motor-M1
// GPIO4, pull low, to motordrv in2, to motor-M2
// GPIO13, pull low, to motordrv in3, to motor-M3
// GPIO12, pull low, to motordrv in4, to motor-M4
// problem: when motor heavy loaded, the voltage can not drop.
// problem: GPIO0 pull-low can not flash, must connect low.
// board S2.


// 20210224 update/new board S2-1
// pins: TX/GND
// pins: RX/GND
// pins: GPIO0/GND
// connector: power +/-, blue, source.
// connector: power +/-, white body, to motor driver board power source. preceded with current limited resistor.
// ADC input: power source 16V:0.892V; 5V:0.257V
// diff to S2-0 version: current limited resistor(for sensing) now imposed between the 2 power sources rather than sensing at the output of mdrv.
// test1: 15V power source, w/o mdrv, 0.03A/0.45Wmax. with mdrv, 23-Ohm resistor climit, motor free run, 0.13A/1.95Amax.
// test2: 15V0.37Amax with 33-Ohm.
// test3: 16V0.5Amax with 30-Ohm.
//
// 20210314 update board S2-1, main chip now changed to ESP-07S.
// resistor now uses 24-Ohm. both 30, 24 are usable.
// flash freq changed to 40MHz.
// sometimes powering-up causes malfunction and consuming high current.
// IDE: Generic ESP8266 module, 115200bps, crystal 26MHz, CPU 160MHz, 4MB flash w/i 40MHz.


// 20210314 add new board S2-2, main chip now changed to ESP-07S.
// resistor now uses 24-Ohm. both 30, 24 are usable.
// sometimes powering-up causes malfunction and consuming high current.
// IDE: Generic ESP8266 module, 115200bps, crystal 26MHz, CPU 160MHz, 4MB flash w/i 40MHz.
// pins: TX/GND
// pins: RX/GND
// pins: GPIO0/GND
// connector: power +/-, blue, source.
// connector: blue, resistor.
// connector: power +/-, white body, be motor driver board power source. preceded with current limited resistor.
// ADC input: power source 16V:0.892V.
// test1: 16V power source, w/i mdrv, 0.028A~0.045A; generally 0.038A. with mdrv, 24-Ohm resistor climit, motor blocked run, 0.555A.


// software:
// 20210226 add fast mode. switched by fast_mode=0/1 which done before Start() and after Stop().
// 20210302 add GPIO16 in 191us period to trigger GPIO14 INT ISR for ADC sampling.
// 20210303 fulfill the voltage inc/dec transition capture.
// 20210314 finished ProbeParams() to generate motor parameters.

// 2021.04.09. update.
// version 0.1:
// board S2-1.
// everything almost ready
#define VERSION_202104090001 "[[202104090001][STA][ESP_TESTAP:87654321][esp8266-iot-motor:80][esp8266-iot-internet][upd:a.bin]"\
            "[MDNS:RT][ESP07S-S2-1][115200:160:26:4M:3MFS:433904SK:80M]]" // so the version number is the 3rd byte, length 12.




#define M1_GPIO             5
#define M2_GPIO             4
#define M3_GPIO             13
#define M4_GPIO             12
#define NEW_PARKING_GPIO    3
#define LED_GPIO            2
#define INT_GPIO            14
#define TRG_GPIO            16

#define MAX_VOLT            16.0f // 16V:0.892V; 5V:0.257V
#define MAX_VOLT_ADC        0.892f
#define REF_VOLT_RATE       (MAX_VOLT/MAX_VOLT_ADC/1024.0f)
#define MPERIOD             5 // 5ms
#define MPERIOD_US          5000 // 5ms
#define MDELAY              delay(MPERIOD*2)
#define SAMPLING_PERIOD     (MPERIOD/2) // 2.5ms, adc.
#define ADC_S_PERIOD        191 // 191us
// there are 2 independent adc sampling, one is 2.5ms once demand, another is 191us periodically but do once demand.
#define DELAY_ADC           delay(SAMPLING_PERIOD)
#define GET_VOLT(x)         {(x)=analogRead(A0); DELAY_ADC;}
#define MAX_TRIP_TIME       2500 // 2.5s
#define DC_PROBING          99 // 99%
#define DELAY_MREST()       delay((MAX_TRIP_TIME)/4)
#define DELAY_MSTBL()       delay((MAX_TRIP_TIME)/8)
#define DELAY_MLONG()       delay((MAX_TRIP_TIME)/2)
#define MSTAGE              7 // ---- test, fixed stage 7
#define DBN                 "/database"
#define FACDEFAULT          "/facdefault"
#define DEFAULT_ADC_DROP    300 // around 5v

// now add fast mode as to solely control GPIO for fast response as well as timing accuracy.
#define M1_sEN              pinMode(M1_GPIO, OUTPUT)
#define M1_cEN              pinMode(M1_GPIO, INPUT)
#define M2_sEN              pinMode(M2_GPIO, OUTPUT)
#define M2_cEN              pinMode(M2_GPIO, INPUT)
#define M3_sEN              pinMode(M3_GPIO, OUTPUT)
#define M3_cEN              pinMode(M3_GPIO, INPUT)
#define M4_sEN              pinMode(M4_GPIO, OUTPUT)
#define M4_cEN              pinMode(M4_GPIO, INPUT)

typedef enum{
    minit=0,
    mstart=1,
    mforward=2,
    mreward=3,
    mforward_pause=4,
    mreward_pause=5,
    mstop=6,
    fstart=1,
    fforward=7,
    freward=8,
    fforward_pause=9,
    freward_pause=10,
    fstop=6,
} eTheState;

unsigned the_dc=0;
unsigned fast_mode=1;

eTheState the_state;
int is_params_valid=0; // if valid and for use
unsigned gen_params=0; // to do the probing
unsigned mtr_works=0; // to bypass the mtr working
unsigned mtr_test=0; // do the test
unsigned rt_sync_the_speed=0; // sync the speed
unsigned rt_the_speed=0; // the speed http got
unsigned RunTimes=0; // intend for renew
String web_link; // website link for get speed
int current_id=-1; // the id currently used
unsigned lock_id=0; // to not change params when http demand
unsigned current_voltage=0; // the rt current voltage

cMultiPwms M1(M1_GPIO, MPERIOD_US, 100, 0, 0);
cMultiPwms M2(M2_GPIO, MPERIOD_US, 100, 0, 0);
cMultiPwms M3(M3_GPIO, MPERIOD_US, 100, 0, 0);
cMultiPwms M4(M4_GPIO, MPERIOD_US, 100, 0, 0);

cMultiPwms LEDB(LED_GPIO, 25000, 500, 1, 1);

cHwTimer timer_get_adc_manip;
IRAM_ATTR unsigned timer_trigger=0;
IRAM_ATTR void isr_get_adc_manip_timer(){digitalWrite(TRG_GPIO, timer_trigger^=1);}

void (*Forward)()=Forward0; // use Forward rather than Forward0 hereafter
void (*Reward)()=Reward0; // use Reward rather than Reward0 hereafter
void (*MFPTR)()=Forward0; // only not this one; do not use MFPTR=Forward because Forward changes and leaves MFPTR alone.
unsigned mtr_reverse=0; // for runtime reversal
#define TR_MFPTR ((MFPTR==Forward)? (MFPTR=Reward): (MFPTR=Forward))


void setup_cmultipwms() {
    //ADC_MODE(ADC_TOUT);
    pinMode(NEW_PARKING_GPIO, INPUT); // by the RX0. set it as the parking pin.

    pinMode(LED_GPIO, OUTPUT); // led
    LEDB.setDC(0); // set as low as on
    LEDB.Resume();

    pinMode(TRG_GPIO, OUTPUT); // this gpio is to trigger intr gpio
    digitalWrite(TRG_GPIO, LOW);
    attachInterrupt(digitalPinToInterrupt(INT_GPIO), isr_get_adc_manip, CHANGE);

    timer_get_adc_manip.setTimer(ADC_S_PERIOD, isr_get_adc_manip_timer);

    // decide mtr reverse or not
    cDataBase sdb(FACDEFAULT);
    sdb.Open();
    if (!sdb.ReadWords(mtr_reverse, "mtr_reverse")) sdb.WriteWords("mtr_reverse", 0, mtr_reverse);
    if (!sdb.ReadWords(fast_mode, "fast_mode")) sdb.WriteWords("fast_mode", 0, fast_mode);
    if (mtr_reverse){
        Forward=Reward0;
        Reward=Forward0;
        MFPTR=Reward0;
    }
    sdb.ReadWords(RunTimes, "RunTimes");
    sdb.Close();
    sdb.ReplaceWordsEx("RunTimes", 0, ++RunTimes);
    if (!(RunTimes%10)){sdb.Renew(); cDataBase(DBN).Renew();}
}


void SetDC_all_temp(unsigned a){
    M1.setDC(a);
    M2.setDC(a);
    M3.setDC(a);
    M4.setDC(a);
}

void SetDC_all(unsigned a){
    SetDC_all_temp(a);
    the_dc=a;
}

void SetDC_temp(class cMultiPwms &m, unsigned a){ // why need to use "class"
    m.setDC(a);
}

void SetDC(cMultiPwms &m, unsigned a){
    SetDC_temp(m, a);
    the_dc=a;
}

void Start(){
    switch (the_state){
        case minit:
        case mstart:
        case mstop:
            digitalWrite(M1_GPIO, LOW);
            digitalWrite(M2_GPIO, LOW);
            digitalWrite(M3_GPIO, LOW);
            digitalWrite(M4_GPIO, LOW);

            pinMode(M1_GPIO, OUTPUT); digitalWrite(M1_GPIO, LOW);
            pinMode(M2_GPIO, OUTPUT); digitalWrite(M2_GPIO, LOW);
            pinMode(M3_GPIO, OUTPUT); digitalWrite(M3_GPIO, LOW);
            pinMode(M4_GPIO, OUTPUT); digitalWrite(M4_GPIO, LOW);

            the_state=mstart;
            SetDC_all(0); // this is must
            cMultiPwms::SyncStart(M1);
            cMultiPwms::SyncNext(M2, 0);
            cMultiPwms::SyncNext(M3, 0);
            cMultiPwms::SyncNext(M4, 0);
            cMultiPwms::SyncEnd(M1);
            SetDC_all(0);
            M1.Pause();
            M2.Pause();
            M3.Pause();
            M4.Pause();

            if (fast_mode){
                M1_cEN;
                M1.Resume();
                M2_cEN;
                M2.Resume();
                M3_cEN;
                M3.Resume();
                M4_cEN;
                M4.Resume();
                the_state=fstart;
            }
            break;
    }
}

void Stop(){
    switch (the_state){
        case mforward:
            M4.Stop();
            MDELAY;
            SetDC_temp(M2, 100);
            SetDC_temp(M1, 100);
            MDELAY;
            M2.Resume();
            MDELAY;
            M1.Stop();
            SetDC(M2, the_dc);
            SetDC(M1, the_dc);
            MDELAY;
            M2.Stop();
            MDELAY;

            M3.Stop();
            MDELAY;
            break;
        case mreward:
            M2.Stop();
            MDELAY;
            SetDC_temp(M4, 100);
            SetDC_temp(M3, 100);
            MDELAY;
            M4.Resume();
            MDELAY;
            M3.Stop();
            SetDC(M4, the_dc);
            SetDC(M3, the_dc);
            MDELAY;
            M4.Stop();
            MDELAY;

            M1.Stop();
            MDELAY;
            break;
        case mforward_pause:
            M1.Stop();
            MDELAY;
            M2.Stop();
            MDELAY;

            M3.Stop();
            MDELAY;
            M4.Stop();
            MDELAY;
            break;
        case mreward_pause:
            M3.Stop();
            MDELAY;
            M4.Stop();
            MDELAY;

            M1.Stop();
            MDELAY;
            M2.Stop();
            MDELAY;
            break;
        case mstart: mstop:
        case fforward: case freward: case fforward_pause: case freward_pause:
            M1.Stop();
            M2.Stop();
            M3.Stop();
            M4.Stop();
            break;
        default: return;
    }
    SetDC_all(0);
    the_state=mstop;

    digitalWrite(M1_GPIO, LOW); pinMode(M1_GPIO, INPUT);
    digitalWrite(M2_GPIO, LOW); pinMode(M2_GPIO, INPUT);
    digitalWrite(M3_GPIO, LOW); pinMode(M3_GPIO, INPUT);
    digitalWrite(M4_GPIO, LOW); pinMode(M4_GPIO, INPUT);

    if (fast_mode) the_state=fstop;
}

void Pause(){
    if (!fast_mode) switch (the_state){
        case mforward:
            the_state=mforward_pause;
            M4.Pause();
            MDELAY;
            SetDC_temp(M2, 100);
            SetDC_temp(M1, 100);
            MDELAY;
            M2.Resume();
            MDELAY;
            SetDC(M2, the_dc);
            SetDC(M1, the_dc);
            break;
        case mreward:
            the_state=mreward_pause;
            M2.Pause();
            MDELAY;
            SetDC_temp(M4, 100);
            SetDC_temp(M3, 100);
            MDELAY;
            M4.Resume();
            MDELAY;
            SetDC(M4, the_dc);
            SetDC(M3, the_dc);
            break;
    }
    else switch (the_state){
        case fforward:
            the_state=fforward_pause;
            M4_cEN;
            M2_sEN;
            break;
        case freward:
            the_state=freward_pause;
            M2_cEN;
            M4_sEN;
            break;
    }
}

void Resume(){
    if (!fast_mode) switch (the_state){
        case mforward_pause:
            the_state=mforward;
            M2.Pause();
            MDELAY;
            M4.Resume();
            MDELAY;
            break;
        case mreward_pause:
            the_state=mreward;
            M4.Pause();
            MDELAY;
            M2.Resume();
            MDELAY;
            break;
    }
    else switch (the_state){
        case fforward_pause:
            the_state=fforward;
            M2_cEN;
            M4_sEN;
            break;
        case freward_pause:
            the_state=freward;
            M4_cEN;
            M2_sEN;
            break;
    }
}

void Forward0(){
    if (!fast_mode) switch (the_state){
        case mstart:
            the_state=mforward;
            M1.Resume();
            MDELAY;
            M4.Resume();
            MDELAY;
            break;
        case mreward:
            the_state=mforward;
            M2.Pause();
            MDELAY;
            SetDC_temp(M4, 100);
            SetDC_temp(M3, 100);
            MDELAY;
            M4.Resume();
            MDELAY;
            M3.Pause();
            SetDC(M4, the_dc);
            SetDC(M3, the_dc);
            MDELAY;
            M1.Resume();
            MDELAY;
            break;
        case mforward_pause:
            the_state=mforward;
            M2.Pause();
            MDELAY;
            M4.Resume();
            MDELAY;
            break;
        case mreward_pause:
            the_state=mforward;
            M3.Pause();
            MDELAY;
            M1.Resume();
            MDELAY;
            break;
    }
    else switch (the_state){
        case fstart:
            the_state=fforward;
            M1_sEN;
            M4_sEN;
            break;
        case freward:
            the_state=fforward;
            M2_cEN;
            M4_sEN;
            M3_cEN;
            M1_sEN;
            break;
        case fforward_pause:
            the_state=fforward;
            M2_cEN;
            M4_sEN;
            break;
        case freward_pause:
            the_state=fforward;
            M3_cEN;
            M1_sEN;
            break;
    }
}

void Reward0(){
    if (!fast_mode) switch (the_state){
        case mstart:
            the_state=mreward;
            M3.Resume();
            MDELAY;
            M2.Resume();
            MDELAY;
            break;
        case mforward:
            the_state=mreward;
            M4.Pause();
            MDELAY;
            SetDC_temp(M2, 100);
            SetDC_temp(M1, 100);
            MDELAY;
            M2.Resume();
            MDELAY;
            M1.Pause();
            SetDC(M2, the_dc);
            SetDC(M1, the_dc);
            MDELAY;
            M3.Resume();
            MDELAY;
            break;
        case mforward_pause:
            the_state=mreward;
            M1.Pause();
            MDELAY;
            M3.Resume();
            MDELAY;
            break;
        case mreward_pause:
            the_state=mreward;
            M4.Pause();
            MDELAY;
            M2.Resume();
            MDELAY;
            break;
    }
    else switch (the_state){
        case fstart:
            the_state=freward;
            M3_sEN;
            M2_sEN;
            break;
        case fforward:
            the_state=freward;
            M4_cEN;
            M2_sEN;
            M1_cEN;
            M3_sEN;
            break;
        case fforward_pause:
            the_state=freward;
            M1_cEN;
            M3_sEN;
            break;
        case freward_pause:
            the_state=freward;
            M4_cEN;
            M2_sEN;
            break;
    }
}

bool isRunning(){
    return ((the_state==mforward) || (the_state==mreward) || (the_state==fforward) || (the_state==freward));
}

void LedBlinking(){
    static unsigned tmos=micros();
    if (cMultiPwms::Timeout(100, tmos)){
        static unsigned cntt=0;
        cntt++;
        switch (the_state){
            case mstart:
                LEDB.detachGpio();
                digitalWrite(LED_GPIO, LOW);
                break;
            case mforward: case fforward:
                LEDB.detachGpio();
                if (cntt&0x4) digitalWrite(LED_GPIO, LOW);
                else digitalWrite(LED_GPIO, HIGH);
                break;
            case mreward: case freward:
                LEDB.detachGpio();
                if (cntt&0x1) digitalWrite(LED_GPIO, LOW);
                else digitalWrite(LED_GPIO, HIGH);
                break;
            case mforward_pause: case fforward_pause:
                LEDB.setGpio(LED_GPIO);
                {unsigned light1=cntt%7+1;
                light1=light1*light1*light1;
                LEDB.setDC(0, light1);}
                break;
            case mreward_pause: case freward_pause:
                LEDB.setGpio(LED_GPIO);
                {unsigned light2=cntt%7+1;
                light2=light2*light2*light2;
                LEDB.setDC(0, 500/light2);}
                break;
            case mstop:
                LEDB.detachGpio();
                digitalWrite(LED_GPIO, HIGH);
                break;
        }
    }
}

void PrintState(){
    static unsigned tmo=micros();
    if (!cMultiPwms::Timeout(3000, tmo)) return;

    switch (the_state){
        case mstart:
            printf("\r\nStart\r\n");
            break;
        case mforward: case fforward:
            printf("\r\nForward\r\n");
            break;
        case mreward: case freward:
            printf("\r\nReward\r\n");
            break;
        case mforward_pause: case mreward_pause: case fforward_pause: case freward_pause:
            printf("\r\nPause\r\n");
            break;
        case mstop:
            printf("\r\nStop\r\n");
            break;
    }
    printf("ADC[%fV]\r\n", float(analogRead(A0))*REF_VOLT_RATE);
}

unsigned get_adc_data=0; // the current adc value. independently updating once demand.
unsigned get_adc_time_us=0; // how long to do this prob. dec by isr. set nonzero by client to start prob.
unsigned get_adc_timestamp=0; // if event happen, set the timestamp/micros. cleared by client before use.
unsigned get_adc_strict_indec=0; // strictly increasing or decreasing on the demanded interval is required.
unsigned get_adc_v1=2; // set by client the event is v1 transits to v2 in inc/dec manner. v can not <=vbr.
unsigned get_adc_v2=2; // set by client the event is v1 transits to v2 in inc/dec manner. v can not <=vbr.
unsigned get_adc_vbr=1; // tolerance.
unsigned get_adc_got=0;
// for example, get_adc_v1=6, get_adc_v2=10, get_adc_vbr=5, get_adc_time_us=350(but done by 382):
// the event of voltage increasingly transits from 6 to 10 within 350us, if happened within specified time, timestamp set.
// if SMPS=191us, it heavily affects wifi into disappeared. so do the sampling once demand.
// ProbAdcEventTrigger(w, x, y, z); while (!ProbAdcEventPolling() && !timeout) delay(1);
ICACHE_RAM_ATTR void isr_get_adc_manip(){
    ////get_adc_data=analogRead(A0);
    if (get_adc_time_us>=ADC_S_PERIOD){
        get_adc_data=analogRead(A0);////
        get_adc_time_us-=ADC_S_PERIOD;
        if (!get_adc_got){
            if ((get_adc_v1<=get_adc_v2) && (get_adc_data<=(get_adc_v1+get_adc_vbr))) get_adc_got=(get_adc_data-get_adc_vbr);
            else if ((get_adc_v1>get_adc_v2) && (get_adc_data>=(get_adc_v1-get_adc_vbr))) get_adc_got=(get_adc_data+get_adc_vbr);
            else if (get_adc_strict_indec) get_adc_time_us=0;
        }
        else if (get_adc_v1<=get_adc_v2){
            if (get_adc_data<(get_adc_v2-get_adc_vbr)){
                if (get_adc_data>=(get_adc_got-get_adc_vbr)) get_adc_got=(get_adc_data-get_adc_vbr);
                else {get_adc_got=0; if (get_adc_strict_indec) get_adc_time_us=0;}
            }
            else {get_adc_timestamp=micros(); get_adc_got=0; get_adc_time_us=0;}
        }
        else {
            if (get_adc_data>(get_adc_v2+get_adc_vbr)){
                if (get_adc_data<=(get_adc_got+get_adc_vbr)) get_adc_got=(get_adc_data+get_adc_vbr);
                else {get_adc_got=0; if (get_adc_strict_indec) get_adc_time_us=0;}
            }
            else {get_adc_timestamp=micros(); get_adc_got=0; get_adc_time_us=0;}
        }
    }
    else if (get_adc_got){
        get_adc_data=analogRead(A0);////
        if ((get_adc_v1<=get_adc_v2) && (get_adc_data>=(get_adc_v2-get_adc_vbr)) ||\
            (get_adc_v1>get_adc_v2) && (get_adc_data<=(get_adc_v2+get_adc_vbr)))
            get_adc_timestamp=micros();
        get_adc_got=0; get_adc_time_us=0;
    }
    else get_adc_time_us=0;
}

void ProbAdcEventTrigger(unsigned v1, unsigned v2, unsigned tolerance, unsigned timespan_us, bool strict_indec=false){
    get_adc_time_us=0;
    get_adc_got=0;
    get_adc_time_us=0;

    get_adc_strict_indec=strict_indec;
    get_adc_v1=v1;
    get_adc_v2=v2;
    get_adc_vbr=tolerance;
    get_adc_timestamp=0;
    get_adc_time_us=timespan_us;
}

unsigned ProbAdcEventPolling(){
    return get_adc_timestamp;
}

unsigned Vidle, Vpause, Vblked0, Vblked1, Vblked2, Vbr0, Vbr, Vrun, Vrunal;
unsigned Rtime_base, Rtime_looser, Rtime_loose, Rtime_tight, Rtime_tightest, Rtime_spring;
unsigned Revo_org_steps, Revo_startup_time, Revo_timespan, Revo_factor;
bool ProbeParams(unsigned course_time_ms){ // about 2 minutes.
    // course_time: time spent for traveling the entire course. since it depends on speed. maximum should be taken(in lowest speed).
    // the course: idle -> start run -> stable run -> meet obstacle -> blocked -> pause
    // with the corresponding state designated: Vidle -> Vrunal -> Vrun -> Vblkal -> Vblked -> Vpause
    // this function is to decide these V states, in addition the voltage drifting should be decided/Vbr.
    // Vrunal should be smallest and is nearly Vblked and is dependent on voltage source, motor speed and sampling speed.
    // Vblked0, tightest/lowest. Vblked1, lowerbound threshold for deciding the trip of the timespan boundary.
    // Vblked2, regarded as just about to met boundary. Vblked 0<1<2.

    unsigned Vmin=1024, Vmax=0, v=0, i=0, j=0, Rtime=0, tmo1=0, tmo2=0;
    Vidle=0, Vpause=0, Vblked0=0, Vblked1=0, Vblked2=0, Vbr0=0, Vbr=0, Vrun=0, Vrunal=1024;
    Rtime_base=0, Rtime_looser=0, Rtime_loose=0, Rtime_tight=0, Rtime_tightest=0, Rtime_spring=0;

    is_params_valid=0;
    mtr_works=0;

    // ---- decide Vidle, Vbr0: stop -> SAMPLING_PERIOD, 100cnt -> start -> forward -> stop -> SAMPLING_PERIOD, 100cnt
    Stop();
    DELAY_MREST();
    while (i++<100){
        GET_VOLT(v);
        Vidle+=v;
        if (v<Vmin) Vmin=v;
        else if (v>Vmax) Vmax=v;
        delay(15); // span 1.5s at all for sensing.
    }

    if (Vidle/100<50){printf("Failed to start probing.\r\n"); return false;}

    if (Vmax>Vmin) Vbr0=Vmax-Vmin;
    else Vbr0=1024;
    Vmin=1024; Vmax=0;
    Start();
    SetDC_all(DC_PROBING);
    MFPTR=Forward;
    MFPTR(); // Forward
    delay(course_time_ms);
    Stop();
    DELAY_MREST();
    while (--i){
        GET_VOLT(v);
        Vidle+=v;
        if (v<Vmin) Vmin=v;
        else if (v>Vmax) Vmax=v;
        delay(15); // span 1.5s at all for sensing.
    }
    if (Vmax<Vmin) Vbr0=1024;
    else if ((Vmax-Vmin)>Vbr0) Vbr0=Vmax-Vmin;
    Vidle/=200;
    printf("[Vidle = (%d, %fV) Vmin = (%d, %fV) Vmax = (%d, %fV)]\r\n",\
        Vidle, Vidle*REF_VOLT_RATE, Vmin, Vmin*REF_VOLT_RATE, Vmax, Vmax*REF_VOLT_RATE);

    // ---- decide Vblked0, Vbr0, Vbr: forward(get) -> reward(get) -> forward(get) -> pause
    Start();
    SetDC_all(DC_PROBING);
    MFPTR(); // Forward
    DELAY_MSTBL(); // time for running stable.
    Vmin=1024; Vmax=0;
    tmo1=micros();
    while (!cMultiPwms::Timeout(course_time_ms, tmo1, true)){
        GET_VOLT(v);
        if (v<Vmin) Vmin=v;
        else if (v>Vmax) Vmax=v;
    }
    TR_MFPTR(); // Reward
    Vblked0=Vmin;
    printf("\r\nVblked0F[%d] ", Vblked0);
    DELAY_MSTBL(); // time for running stable.
    Vmin=1024; Vmax=0;
    tmo1=micros();
    while (!cMultiPwms::Timeout(course_time_ms, tmo1, true)){
        GET_VOLT(v);
        if (v<Vmin) Vmin=v;
        else if (v>Vmax) Vmax=v;
    }
    Pause();
    DELAY_MREST();
    TR_MFPTR(); // Forward
    DELAY_MSTBL(); // time for running stable.
    if (Vmin>Vblked0) Vblked0=Vmin;
    printf("Vblked0R[%d] ", Vblked0);
    Vmin=1024; Vmax=0;
    tmo1=micros();
    while (!cMultiPwms::Timeout(course_time_ms, tmo1, true)){
        GET_VOLT(v);
        if (v<Vmin) Vmin=v;
        else if (v>Vmax) Vmax=v;
    }
    Pause();
    DELAY_MREST();
    if (Vmin>Vblked0) Vblked0=Vmin;
    printf("Vblked0F[%d]\r\n", Vmin);

    if (Vblked0>=Vidle && Vbr0!=1024) Vblked0-=(Vbr0*2);
    printf("[take Vmin for Vblked0. Vblked0 = (%d, %fV) Vmin = (%d, %fV) Vmax = (%d, %fV)]\r\n",\
        Vblked0, Vblked0*REF_VOLT_RATE, Vmin, Vmin*REF_VOLT_RATE, Vmax, Vmax*REF_VOLT_RATE);

    // ---- decide Vpause, Vbr0, Vbr, Vblked0.
    Vmin=1024; Vmax=0;
    tmo1=micros();
    while (!cMultiPwms::Timeout(course_time_ms, tmo1, true)){
        GET_VOLT(v);
        if (v<Vmin) Vmin=v;
        else if (v>Vmax) Vmax=v;
    }
    Vpause=(Vmin+Vmax)/2;
    if (Vmax<Vmin) Vpause=0;
    else if ((Vmax-Vmin)>Vbr0) Vbr0=Vmax-Vmin;
    if (Vbr0>20){
        printf("Vbr0 too high, it caused by low voltage. low voltage could cause malfunctions.\r\n");
        return false;
    }
    Vblked0+=Vbr0;
    Vbr=Vbr0*15; // might be more ok. // <------------------------------------------- adjust from here!!!

    printf("[Vpause = (%d, %fV) Vmin = (%d, %fV) Vmax = (%d, %fV)]\r\n",\
        Vpause, Vpause*REF_VOLT_RATE, Vmin, Vmin*REF_VOLT_RATE, Vmax, Vmax*REF_VOLT_RATE);
    printf("[Vbr = (%d, %fV) Vbr0 = (%d, %fV) Vblked0 = (%d, %fV)]\r\n",\
        Vbr, Vbr*REF_VOLT_RATE, Vbr0, Vbr0*REF_VOLT_RATE, Vblked0, Vblked0*REF_VOLT_RATE);

    // ---- reward -> pause. decide Vblked1, Vblked2 and Vrunal, Vrunal should be the smallest: (Vrunal...Vblked...Vrun...Vpause...Vidle)
    // since there is a delay right after forward/reward, Vrunal might not get the minimum one.
    Vmin=Vidle;
    TR_MFPTR; // into reward
    tmo1=micros();
    tmo2=tmo1;
    while (!cMultiPwms::Timeout(course_time_ms/2, tmo1, true)){
        if (cMultiPwms::Timeout(SAMPLING_PERIOD, tmo2, true)) MFPTR();
        GET_VOLT(v);
        if (v<Vmin) Vmin=v;
    }
    Pause();
    DELAY_MREST();
    tmo1=micros();
    tmo2=tmo1;
    while (!cMultiPwms::Timeout(course_time_ms/4, tmo1, true)){
        if (cMultiPwms::Timeout(SAMPLING_PERIOD, tmo2, true)) MFPTR();
        GET_VOLT(v);
        if (v<Vmin) Vmin=v;
    }
    Pause();
    DELAY_MREST();
    Vrunal=Vmin;
    printf("[Vrunal = (%d, %fV)]\r\n", Vrunal, Vrunal*REF_VOLT_RATE);

    // next, assign Vblked1, Vblked2 and rearrange Vrunal.
    if (Vblked0<Vrunal) Vrunal=Vblked0;
    Vblked1=Vblked2=Vblked0;

    // ---- decide Vrun. while run, Vrun is the maximum.
    Vmax=0; i=6;
    while (--i){
        TR_MFPTR();
        tmo1=micros();
        while (!cMultiPwms::Timeout(course_time_ms, tmo1, true)){
            GET_VOLT(v);
            if (v>Vmax) Vmax=v;
        }
    }
    Pause();
    DELAY_MREST();
    Vrun=Vmax;

    // ---- refine again: Vrunal ... Vblked-> ... <-Vrun
    Vrun=Vrun-Vbr; // <------------------------------------------- adjust from here!!!
    Vblked2=(Vblked0+2*Vrun)/3; // <------------------------------------------- adjust from here!!!
    Vblked1=(2*Vblked0+Vblked2)/3; // <------------------------------------------- adjust from here!!!
    /////Vblked2=(Vblked1+6*Vblked2)/7; // <------------------------------------------- adjust from here!!!
    Vrunal=(2*Vrunal+Vblked0)/3; // <------------------------------------------- adjust from here!!!

    printf("[Vrunal = (%d, %fV) Vblked0 = (%d, %fV) Vblked1 = (%d, %fV) Vblked2 = (%d, %fV) Vrun = (%d, %fV)]\r\n",\
        Vrunal, Vrunal*REF_VOLT_RATE, Vblked0, Vblked0*REF_VOLT_RATE, Vblked1, Vblked1*REF_VOLT_RATE, Vblked2, Vblked2*REF_VOLT_RATE,\
        Vrun, Vrun*REF_VOLT_RATE);
    printf("\r\n[Vidle, Vpause, Vblked0, Vblked1, Vblked2, Vbr0, Vbr, Vrun, Vrunal = (%d, %d, %d, %d, %d, %d, %d, %d, %d)]\r\n",\
    Vidle, Vpause, Vblked0, Vblked1, Vblked2, Vbr0, Vbr, Vrun, Vrunal);

    // ---- yet another new last trial. we use periodically adc sampling to probe timeline of Vrunal, Vblkal.
    // Vblked1 -> Vrunal -> Vblked1 -> Vblked2
    Vmin=5;
    Vmax=0;
    while (--Vmin){
        delay(1000);
        i=0;
        tmo1=micros();
        while (!cMultiPwms::Timeout(course_time_ms+10, tmo1, true, true)){
            if (!i++){
                j=micros();
                ProbAdcEventTrigger(Vblked1, Vrunal, Vbr, 150000); // under 150ms, pre-Vrunal
                TR_MFPTR();
            }
            if (j){
                if (v=ProbAdcEventPolling()){ // hit
                    ProbAdcEventTrigger(Vblked1, Vblked2, Vbr, 150000); // under 150ms, post-Vrunal
                    while (!ProbAdcEventPolling() && (j+150000)>micros()) delay(10); // under 150ms for Vrunal
                    printf("[reward] start timestamp[%d], Vrunal-falling timestamp[%d], rising[%d], diff1[%d], diff2[%d]\r\n",\
                    j, v, ProbAdcEventPolling(), v-j, ProbAdcEventPolling()-v);
                    if (ProbAdcEventPolling() && Vmax<(ProbAdcEventPolling()-j)) Vmax=ProbAdcEventPolling()-j;
                    j=0;
                }
            }
            delay(1);
        }
        i=0;
        tmo1=micros();
        while (!cMultiPwms::Timeout(course_time_ms+10, tmo1, true, true)){
            if (!i++){
                j=micros();
                ProbAdcEventTrigger(Vblked1, Vrunal, Vbr, 150000); // under 150ms, pre-Vrunal
                TR_MFPTR();
            }
            if (j){
                if (v=ProbAdcEventPolling()){ // hit
                    ProbAdcEventTrigger(Vblked1, Vblked2, Vbr, 150000); // under 150ms, post-Vrunal
                    while (!ProbAdcEventPolling() && (j+150000)>micros()) delay(10); // under 150ms for Vrunal
                    printf("[forward] start timestamp[%d], Vrunal-falling timestamp[%d], rising[%d], diff1[%d], diff2[%d]\r\n",\
                    j, v, ProbAdcEventPolling(), v-j, ProbAdcEventPolling()-v);
                    if (ProbAdcEventPolling() && Vmax<(ProbAdcEventPolling()-j)) Vmax=ProbAdcEventPolling()-j;
                    j=0;
                }
            }
            delay(1);
        }
        Pause();
        DELAY_MREST();
    }
    Rtime_base=(Vmax+999)/1000; // Vblked1->Vrunal->Vblked2
    printf("[Rtime_base = (%dus, %dms)]\r\n", Vmax, Rtime_base);
    if (Rtime_base<5) if (Rtime_base<1) return false; else Rtime_base=30; // <------------------------------------------- adjust from here!!!

    // ---- determine a trip time, Rtime_tightest
    i=5;
    Vmax=0;
    while (--i){
        TR_MFPTR();
        tmo1=micros();
        tmo2=tmo1;
        delay(Rtime_base); // since Vrunal is the smallest V which could affect the result into wrong. this is also a dependent one!!!
        while (!cMultiPwms::Timeout(course_time_ms, tmo1, true)){
            GET_VOLT(v);
            if (v<Vblked1){
                GET_VOLT(v);
                if (v<Vblked1){
                    Vmin=v;
                    GET_VOLT(v);
                    if (v<Vmin) break;
                }
            }
        }
        tmo2=micros()-tmo2;
        printf("Rtime[%d]\r\n", tmo2);
        if (tmo2>Vmax) Vmax=tmo2;
        Rtime+=tmo2;
    }
    Pause();
    DELAY_MREST();
    Rtime/=4;
    Rtime_tightest=Rtime;
    printf("Rtime_tightest[%d]\r\n", Rtime_tightest);
    if (Vmax>(Rtime_tightest+Rtime_base*10000)) return false; // <------------------------------------------- adjust from here!!!

    // ---- determine a trip time, for Rtime_tight, Rtime_loose, Rtime_looser
    #define DECSEVEN 7
    unsigned f=0, k=0, p=0, q[DECSEVEN]={0, 0, 0, 0, 0, 0, 0};
    unsigned r[DECSEVEN]={9, 4, 6, 2, 4, 3, 2}; // decrements
    unsigned s=1; // for escape if no blocked.

    for (; p<DECSEVEN && (s || p<3); p++){ // Rtime_tight=q[0], Rtime_loose=q[1], Rtime_looser=q[2].
        s=0;
        i=5;
        j=Rtime_base+Rtime/1000; // us to ms
        Vmax=0;
        Vmin=1;
        while (--i){
            DELAY_MSTBL(); // put it into stable.
            if (Vmin) Vmax++;
            Vmin=0;
            TR_MFPTR();
            if (j>(Vmax*r[p]+Rtime_base)){
                delay(Rtime_base); // startup time ensured.
                k=j-Vmax*r[p]-Rtime_base;
            }
            else k=j;
            tmo1=micros();
            tmo2=tmo1;
            printf("\r\nOrg[%d], Shrink[%d]", j, k);
            while (!cMultiPwms::Timeout(k, tmo1, true)){
                GET_VOLT(v);
                if (v<Vblked1){
                    Vmin=v;
                    GET_VOLT(v);
                    if (v<Vmin){
                        Vmin=v;
                        GET_VOLT(v);
                        if (v<Vmin-Vbr0) break;
                        else Vmin=0;
                    }
                    else Vmin=0;
                }
                else Vmin=0;
            }
            tmo2=micros()-tmo2;
            printf("\r\n\r\nRtime%d[%d]\r\n", p+1, tmo2);
            Rtime+=tmo2;
            if (Vmin) s=1;
            else if (!Rtime_spring) Rtime_spring=p+1; // the 1st no tension. p+1 has no reason purpose.
        }
        Rtime/=5;
        q[p]=Rtime;
        if (Rtime_spring && (Rtime_spring==(p+1))) Rtime_spring=(Rtime_tightest-q[p])/2;
    }
    printf("Rtime_tightest[%d]\r\n", Rtime_tightest);
    printf("Rtime_tight[%d]\r\n", (Rtime_tight=q[0]));
    printf("Rtime_loose[%d]\r\n", (Rtime_loose=q[1]));
    printf("Rtime_looser[%d]\r\n", (Rtime_looser=q[2]));
    printf("Rtime_spring[%d]\r\n", Rtime_spring);

    Rtime_tightest=Rtime_looser+Rtime_spring; // <------------------------------------------- adjust from here!!!

    // ---- determine the runnable minimum time, for generating stages and timespan.
    // since motors have volatile startup voltage/current which depending on some factors so affect it what revolve.
    // we only take this as a single step time and discard others that matter about.
    k=0;
    Vmin=Rtime_base+1;
    Vmax=Rtime_tightest/1000;
    MFPTR();
    delay(Vmax);
    Pause();
    delay(10);
    s=Vmax/Vmin;
    printf("max stages=%d %d %d\r\n", Vmax, Vmin, s);
    if (s>22) s=22;
    for (; s; s--){
        printf("current stages=%d\r\n", s);
        Vmin=Rtime_base;
        Vmax=Rtime_tightest/1000;
        while(1){
            p=(Vmin+Vmax)/2;
            if (p==Vmin){
                if (!k){Vmin=s; s=1;} // terminated
                else k=0;
                break;
            }
            TR_MFPTR;
            i=0;
            while (1){
                printf("%s, stage=%d, time=%d.\r\n", ((MFPTR==Reward)? "Reward": "Forward"), i+1, p);
                MFPTR();
                delay(p);
                GET_VOLT(v);
                Pause();
                delay(5);
                if (v<=Vblked1){
                    if ((i+1)<s){
                        k=1;
                        printf("stages %d not fit, too strong/weak\r\n", s);
                    }
                    else k=0;
                    Vmax=p;
                    break;
                }
                if (++i>s){
                    MFPTR();
                    delay(p);
                    GET_VOLT(v);
                    delay(Rtime_tightest/1000-p);
                    Pause();
                    delay(5);
                    Vmin=p;
                    if (v>Vblked1){
                        printf("stages %d not fit, too many\r\n", s);
                        k=1;
                    }
                    else k=0;
                    break;
                }
            }
        }
    }
    printf("\r\nAwarded stage and time are (%d, %d)\r\n", s=Vmin, p);
    Revo_org_steps=s;

    // ---- test, fixed stage MSTAGE 7
    // decide f, s(basetime), p(steptime)
    f=1;
    if (s<=MSTAGE) s=Rtime_base;
    else if (s<(MSTAGE*2)){
        s-=MSTAGE;
        s*=p;
        s/=MSTAGE;
        s+=Rtime_base;
    }
    else {
        f=s/MSTAGE;
        s=s%MSTAGE*p/MSTAGE+Rtime_base;
    }
    Vmax=20;
    MFPTR=Reward;
    MFPTR();
    delay(Rtime_tightest/1000);
    Pause();
    delay(5);
    k=0; // 0=close; MSTAGE=open.
    tmo1=Rtime_tightest/1000;
    while (--Vmax){
        i=j=(random(MSTAGE-1)+1); // 1 to MSTAGE-1
        if (i>k){
            MFPTR=Forward;
            i-=k;
            k=j;
        }
        else if (i<k){
            MFPTR=Reward;
            i=k-i;
            k=j;
        }
        else if (Vmax%2){ // opened
            i=MSTAGE-k;
            k=MSTAGE;
            MFPTR=Forward;
        }
        else { // closed
            i=k;
            k=0;
            MFPTR=Reward;
        }
        printf("\r\n\r\ngpos[%d]\r\n", k);
        delay(750);
        if (!k || (k==MSTAGE)){
            MFPTR();
            delay(s);
            tmo2=micros();
            while (!cMultiPwms::Timeout(tmo1, tmo2, true)){
                GET_VOLT(v);
                if (v<Vblked1){
                    Vmin=Vblked1-Vbr0;
                    GET_VOLT(v);
                    if (v<Vmin){
                        Vmin=v;
                        GET_VOLT(v);
                        if (v<Vmin-Vbr0) break;
                    }
                }
            }
            Pause();
            delay(5);
        }
        else {
            MFPTR();
            delay(s);
            Pause();
            delay(5);
            for (int w=0; w<(i*f); w++){
                MFPTR();
                delay(p);
                Pause();
                delay(5);
            }
        }
    }
    Revo_startup_time=s;
    Revo_timespan=p;
    Revo_factor=f;

    is_params_valid=1;
    mtr_works=1;

    printf("\r\nDone. org stages, startup time, timespan and factor are (%d %d, %d, %d)\r\n",\
        Revo_org_steps, Revo_startup_time, Revo_timespan, Revo_factor);
    return true;
}

void ManuallyControlMtr(unsigned a){ // 0, open; 1, close, 2, at loose close, 3, efficiency, 4, performance, 5, loose open.
    unsigned v, av;
    void (*fp)();
    GET_VOLT(v); delay(100); av=v;
    GET_VOLT(v); delay(100); av+=v;
    GET_VOLT(v); delay(100); av=(av+v)/3;
    if (!a || (a==5)) fp=Forward;
    else fp=Reward;
    fp();
    delay(150);
    unsigned tmo=micros();
    while (!cMultiPwms::Timeout(MAX_TRIP_TIME, tmo, true, true)){
        delay(25);
        GET_VOLT(v);
        if (v+DEFAULT_ADC_DROP<av){Pause(); break;}
    }
    if (a==2){Forward(); delay(125);}
    else if (a==3){Forward(); delay(100); Pause(); Resume(); delay(100); Pause(); Resume(); delay(50);}
    else if (a==4){Forward(); delay(100); Pause(); Resume(); delay(100); Pause(); Resume(); delay(125);}
    else if (a==5){Reward(); delay(125);}
    Pause();
}

void TestParams(unsigned stages, unsigned startup, unsigned span, unsigned factor){
    unsigned x=30; // 30 times about 35 seconds
    unsigned k=0; // 0=close; stages=open.
    unsigned tmo1=Rtime_tightest/1000, tmo2=0;
    unsigned i, j, Vmin, v;

    MFPTR=Reward;
    MFPTR();
    delay(tmo1);
    Pause();
    delay(5);
    while (--x){
        i=j=(random(stages-1)+1); // 1 to stages-1
        if (i>k){
            MFPTR=Forward;
            i-=k;
            k=j;
        }
        else if (i<k){
            MFPTR=Reward;
            i=k-i;
            k=j;
        }
        else if (x%2){ // opened
            i=stages-k;
            k=stages;
            MFPTR=Forward;
        }
        else { // closed
            i=k;
            k=0;
            MFPTR=Reward;
        }
        printf("\r\n\r\ngpos[%d]\r\n", k);
        delay(750);
        if (!k || (k==stages)){
            MFPTR();
            delay(startup);
            tmo2=micros();
            while (!cMultiPwms::Timeout(tmo1, tmo2, true)){
                GET_VOLT(v);
                if (v<Vblked1){
                    Vmin=Vblked1-Vbr0;
                    GET_VOLT(v);
                    if (v<Vmin){
                        Vmin=v;
                        GET_VOLT(v);
                        if (v<Vmin-Vbr0) break;
                    }
                }
            }
            Pause();
            delay(5);
        }
        else {
            MFPTR();
            delay(startup);
            Pause();
            delay(5);
            for (int w=0; w<(i*factor); w++){
                MFPTR();
                delay(span);
                Pause();
                delay(5);
            }
        }
    }
}

void MtrMovement(unsigned stage){ // 0=close; MSTAGE=open.
    static unsigned the_speed=stage;
    unsigned tmo=0;
    unsigned i, j, Vmin, v, w;

    if (the_speed==stage) return;

    i=j=stage;
    if (i>the_speed){
        MFPTR=Forward;
        i-=the_speed;
        the_speed=j;
    }
    else if (i<the_speed){
        MFPTR=Reward;
        i=the_speed-i;
        the_speed=j;
    }
    else if (i>=MSTAGE){ // opened
        i=MSTAGE-the_speed;
        the_speed=MSTAGE;
        MFPTR=Forward;
    }
    else { // closed
        i=the_speed;
        the_speed=0;
        MFPTR=Reward;
    }
    printf("\r\n\r\ngpos[%d]\r\n", the_speed);
    if (!the_speed || (the_speed==MSTAGE)){
        MFPTR();
        delay(Revo_startup_time);
        tmo=micros();
        while (!cMultiPwms::Timeout(Rtime_tightest/1000, tmo, true)){
            GET_VOLT(v);
            if (v<Vblked1){
                Vmin=Vblked1-Vbr0;
                GET_VOLT(v);
                if (v<Vmin){
                    Vmin=v;
                    GET_VOLT(v);
                    if (v<Vmin-Vbr0) break;
                }
            }
        }
        Pause();
        delay(5);
    }
    else {
        MFPTR();
        delay(Revo_startup_time);
        Pause();
        delay(5);
        for (w=0; w<(i*Revo_factor); w++){
            MFPTR();
            delay(Revo_timespan);
            Pause();
            delay(5);
        }
    }
}


void my_http_request(){ // called only if http_request_ip assigned.
    if (rt_sync_the_speed){
        static unsigned tmo=micros();
        if (Timeout(400, tmo)){ // each 0.4 seconds to send http request.
            extern String web_link;
            if (HTTPC_http_client_connect(web_link.c_str())) rt_the_speed=http_got_str.toInt();
        }
    }
}

void my_setup_http_server(){

    uri_list+=\
        "/version\r\n"\
        "/ls\r\n"\
        "/del?filename=<name>\r\n"\
        "/cp?src=<src_filename>&dst=<dst_filename>\r\n"\
        "/reset\r\n"\
        "/factory/loaddefault\r\n"\
        "/factory/casttodefault\r\n"\
        "/factory/format\r\n"\
        "/factory/emptydb\r\n"\
        "/factory/emptyfac\r\n"\
        "/factory/renew\r\n"\
        "/factory/debug/dump\r\n"\
        "/params/mtr/listparams\r\n"\
        "/params/mtr/loadparams\r\n"\
        "/params/mtr/deleteparams\r\n"\
        "/params/mtr/unlockparams\r\n"\
        "/params/mtr/probe\r\n"\
        "/params/mtr/writeback\r\n"\
        "/params/mtr/on\r\n"\
        "/params/mtr/off\r\n"\
        "/params/mtr/syncon\r\n"\
        "/params/mtr/syncoff\r\n"\
        "/params/mtr/test\r\n"\
        "/params/gate/open\r\n"\
        "/params/gate/looseopen\r\n"\
        "/params/gate/close\r\n"\
        "/params/gate/looseclose\r\n"\
        "/params/gate/efficiency\r\n"\
        "/params/gate/performance\r\n"\
        "/factory/mtr/reverse\r\n";

    server_p->on("/version", []() { // show version
        server_p->send(200, "text/plain", VERSION_202104090001);
    });

    server_p->on("/ls", []() { // list directory
        String str="";
        cDataBase::Ls("/", &str);
        server_p->send(200, "text/plain", str.c_str());
    });

    server_p->on("/del", []() { // to delete the specified file.
        String fn=server_p->arg("filename");
        if (fn=="") server_p->send(404, "text/plain", "0. page not found.");
        else {
            if (cDataBase((String("/")+fn).c_str()).Remove())
                server_p->send(200, "text/plain", (String("1. removed: ")+fn).c_str());
            else server_p->send(404, "text/plain", "0. file not found.");
        }
    });

    server_p->on("/cp", []() { // to copy files.
        String src=server_p->arg("src");
        String dst=server_p->arg("dst");
        if (src=="") server_p->send(404, "text/plain", "0. page not found.");
        else if (dst=="") server_p->send(404, "text/plain", "0. page not found.");
        else {
            src="/"+src;
            dst="/"+dst;
            cDataBase::Cp(src.c_str(), dst.c_str());
            server_p->send(200, "text/plain", "1. copied.");
        }
    });

    server_p->on("/reset", []() {
        server_p->send(200, "text/plain", "now reset...");
        Stop();
        delay(10);
        while (1);
    });

    server_p->on("/factory/loaddefault", []() {
        server_p->send(200, "text/plain", "1. factory: default loaded...");
    });

    server_p->on("/factory/casttodefault", []() {
        server_p->send(200, "text/plain", "1. factory: set as default...");
    });

    server_p->on("/factory/format", []() {
        if (cDataBase::Format()) server_p->send(200, "text/plain", "1. factory: now formatted...");
        else server_p->send(200, "text/plain", "0. factory: format failed...");
    });

    server_p->on("/factory/emptydb", []() {
        server_p->send(200, "text/plain", "1. factory: now clear database...");
        cDataBase(DBN).Remove();
    });

    server_p->on("/factory/emptyfac", []() {
        server_p->send(200, "text/plain", "1. factory: now clear factory...");
        cDataBase(FACDEFAULT).Remove();
    });

    server_p->on("/factory/renew", []() {
        server_p->send(200, "text/plain", "1. factory: files renew...");
        cDataBase(FACDEFAULT).Renew();
        cDataBase(DBN).Renew();
    });

    server_p->on("/factory/debug/dump", []() {
        server_p->send(200, "text/plain", "1. factory: default and database dumpped...");
        cDataBase sdb(FACDEFAULT), tdb(DBN);
        sdb.LsRecords(); tdb.LsRecords();
        sdb.DumpAll(0); sdb.DumpAll(1); tdb.DumpAll(0); tdb.DumpAll(1);
    });

    server_p->on("/factory/mtr/reverse", []() {
        server_p->send(200, "text/plain", "1. factory: done...");
        extern unsigned mtr_reverse;
        extern void (*Forward)();
        extern void (*Reward)();
        extern void (*MFPTR)();
        mtr_reverse^=1;
        cDataBase sdb(FACDEFAULT);
        sdb.ReplaceWordsEx("mtr_reverse", 0, mtr_reverse);
        if (mtr_reverse){
            Forward=Reward0;
            Reward=Forward0;
            MFPTR=Reward0;
        }
        else {
            Forward=Forward0;
            Reward=Reward0;
            MFPTR=Forward0;
        }
    });

    server_p->on("/params/mtr/listparams", []() {
        cDataBase tdb(DBN);
        String str="";
        ListParams(&tdb, str);
        server_p->send(200, "text/plain", ("1. mtr parameters:\r\n"+str).c_str());
    });

    server_p->on("/params/mtr/loadparams", []() {
        String id=server_p->arg("id");
        int iid;
        cDataBase tdb(DBN);
        extern bool LoadProbeParams(cDataBase*, int);
        if ((id=="") || (iid=id.toInt())<0 || !LoadProbeParams(&tdb, iid)){
            server_p->send(200, "text/plain", "0. page not found.");
            return;
        }
        lock_id=1;
        server_p->sendHeader("Location", "/params/mtr/listparams");
        server_p->send(303);
    });

    server_p->on("/params/mtr/deleteparams", []() {
        String id=server_p->arg("id");
        int iid;
        if ((id=="") || (iid=id.toInt())<0 || (current_id==iid)){
            server_p->send(200, "text/plain", "0. page not found or id in use.");
            return;
        }
        cDataBase(DBN).DeleteRecords(iid);
        server_p->send(200, "text/plain", "1. done the delete.");
    });

    server_p->on("/params/mtr/unlockparams", []() {
        lock_id=0;
        server_p->send(200, "text/plain", "1. unlocked...");
    });

    server_p->on("/params/mtr/probe", []() {
        server_p->send(200, "text/plain", "1. mtr probing...");
        extern unsigned gen_params;
        gen_params=1;
    });

    server_p->on("/params/mtr/writeback", []() {
        extern bool StoreProbeParams(cDataBase *db, int id=0);
        cDataBase sdb(DBN);
        if (StoreProbeParams(&sdb)) server_p->send(200, "text/plain", "1. params: mtr written...");
        else server_p->send(200, "text/plain", "0. params: mtr written failed...");
    });

    server_p->on("/params/mtr/on", []() { // mtr on
        server_p->send(200, "text/plain", "1. params: mtr on...");
        ManuallyControlMtr(2); // close first
        extern unsigned mtr_works;
        mtr_works=1;
    });

    server_p->on("/params/mtr/off", []() { // mtr off
        server_p->send(200, "text/plain", "1. params: mtr off...");
        ManuallyControlMtr(2); // close first
        extern unsigned mtr_works;
        mtr_works=0;
    });

    server_p->on("/params/mtr/syncon", []() { // sync the speed on
        extern unsigned rt_sync_the_speed;
        rt_sync_the_speed=1;
        unsigned mw=mtr_works;
        server_p->send(200, "text/plain", "1. params: mtr sync speed...");
        cDataBase sdb(DBN);
        sdb.ReplaceWordsEx("rt_sync_the_speed", 0, rt_sync_the_speed);
        ManuallyControlMtr(2); // close first
        mtr_works=mw;
    });

    server_p->on("/params/mtr/syncoff", []() { // sync the speed off
        extern unsigned rt_sync_the_speed;
        rt_sync_the_speed=0;
        unsigned mw=mtr_works;
        server_p->send(200, "text/plain", "1. params: mtr abort speed...");
        cDataBase sdb(DBN);
        sdb.ReplaceWordsEx("rt_sync_the_speed", 0, rt_sync_the_speed);
        ManuallyControlMtr(2); // close first
        mtr_works=mw;
    });

    server_p->on("/params/mtr/test", []() { // sync the speed off
        extern unsigned mtr_test;
        mtr_test=1;
        server_p->send(200, "text/plain", "1. params: mtr test starts...");
    });

    server_p->on("/params/gate/open", []() {
        extern void ManuallyControlMtr(unsigned a);
        server_p->send(200, "text/plain", "1. params: gate opened...");
        ManuallyControlMtr(0);
        mtr_works=0;
    });

    server_p->on("/params/gate/looseopen", []() {
        extern void ManuallyControlMtr(unsigned a);
        server_p->send(200, "text/plain", "1. params: gate opened...");
        ManuallyControlMtr(5);
        mtr_works=0;
    });

    server_p->on("/params/gate/close", []() {
        extern void ManuallyControlMtr(unsigned a);
        server_p->send(200, "text/plain", "1. params: gate closed...");
        ManuallyControlMtr(1);
        mtr_works=0;
    });

    server_p->on("/params/gate/looseclose", []() {
        extern void ManuallyControlMtr(unsigned a);
        server_p->send(200, "text/plain", "1. params: gate looseclosed...");
        ManuallyControlMtr(2);
        mtr_works=0;
    });

    server_p->on("/params/gate/efficiency", []() {
        extern void ManuallyControlMtr(unsigned a);
        server_p->send(200, "text/plain", "1. params: fuel efficiency...");
        ManuallyControlMtr(3);
        mtr_works=0;
    });

    server_p->on("/params/gate/performance", []() {
        extern void ManuallyControlMtr(unsigned a);
        server_p->send(200, "text/plain", "1. params: performance");
        ManuallyControlMtr(4);
        mtr_works=0;
    });
}


bool StoreProbeParams(cDataBase *db, int id=0){
    if (!is_params_valid) return false;
    db->Open();
    if (!id) id=db->FindMaxIdEx("Vidle")+1; // 0/not found, inc/found
    db->WriteWords("Vidle", id, Vidle);
    db->WriteWords("Vpause", id, Vpause);
    db->WriteWords("Vblked0", id, Vblked0);
    db->WriteWords("Vblked1", id, Vblked1);
    db->WriteWords("Vblked2", id, Vblked2);
    db->WriteWords("Vbr0", id, Vbr0);
    db->WriteWords("Vbr", id, Vbr);
    db->WriteWords("Vrun", id, Vrun);
    db->WriteWords("Vrunal", id, Vrunal);
    db->WriteWords("Rtime_base", id, Rtime_base);
    db->WriteWords("Rtime_looser", id, Rtime_looser);
    db->WriteWords("Rtime_loose", id, Rtime_loose);
    db->WriteWords("Rtime_tight", id, Rtime_tight);
    db->WriteWords("Rtime_tightest", id, Rtime_tightest);
    db->WriteWords("Rtime_spring", id, Rtime_spring);
    db->WriteWords("Revo_org_steps", id, Revo_org_steps);
    db->WriteWords("Revo_startup_time", id, Revo_startup_time);
    db->WriteWords("Revo_timespan", id, Revo_timespan);
    db->WriteWords("Revo_factor", id, Revo_factor);
    db->LsRecords();
    db->Close();
    printf("[%d %d %d %d %d]\r\n", Vidle, Vpause, Vblked0, Vblked1, Vblked2);
    printf("[%d %d %d %d %d]\r\n", Vbr0, Vbr, Vrun, Vrunal, Rtime_base);
    printf("[%d %d %d %d %d]\r\n", Rtime_looser, Rtime_loose, Rtime_tight, Rtime_tightest, Rtime_spring);
    printf("[%d %d %d %d id=%d]\r\n", Revo_org_steps, Revo_startup_time, Revo_timespan, Revo_factor, id);
    printf("Done written the parameters to the File. Vidle=%fV\r\n", Vidle*REF_VOLT_RATE);
    current_id=id;
    return true;
}

bool LoadProbeParams(cDataBase *db, int id=-1){
    if (lock_id) return false;
    int maxi;
    unsigned v, av;
    GET_VOLT(v); av=v; // 1
    db->Open();
    if (((maxi=db->FindMaxIdEx("Vidle"))==-1) || maxi<id){db->Close(); return false;}
    GET_VOLT(v); av+=v; // 2
    if (id==-1) id=0;
    else maxi=id;

    unsigned vidles[maxi+1-id];
    for (int i=id; i<=maxi; i++){
        if (!db->ReadWords(vidles[i-id], "Vidle", i)) vidles[i-id]=-1;
        //printf("Vidle[%d, %d]\r\n", i, vidles[i-id]);
    }
    GET_VOLT(v); av=(v+av)/3; // 3

    unsigned mini=((av>vidles[0])? av-vidles[0]: vidles[0]-av);
    v=id;
    for (int i=id; i<=maxi; i++){
        if (vidles[i-id]>av){if (mini>vidles[i-id]-av){mini=vidles[i-id]-av; v=i;}}
        else if (mini>av-vidles[i-id]){mini=av-vidles[i-id]; v=i;}
    }
    Vidle=vidles[v-id];
    id=v;
    db->ReadWords(Vpause, "Vpause", id);
    db->ReadWords(Vblked0, "Vblked0", id);
    db->ReadWords(Vblked1, "Vblked1", id);
    db->ReadWords(Vblked2, "Vblked2", id);
    db->ReadWords(Vbr0, "Vbr0", id);
    db->ReadWords(Vbr, "Vbr", id);
    db->ReadWords(Vrun, "Vrun", id);
    db->ReadWords(Vrunal, "Vrunal", id);
    db->ReadWords(Rtime_base, "Rtime_base", id);
    db->ReadWords(Rtime_looser, "Rtime_looser", id);
    db->ReadWords(Rtime_loose, "Rtime_loose", id);
    db->ReadWords(Rtime_tight, "Rtime_tight", id);
    db->ReadWords(Rtime_tightest, "Rtime_tightest", id);
    db->ReadWords(Rtime_spring, "Rtime_spring", id);
    db->ReadWords(Revo_org_steps, "Revo_org_steps", id);
    db->ReadWords(Revo_startup_time, "Revo_startup_time", id);
    db->ReadWords(Revo_timespan, "Revo_timespan", id);
    db->ReadWords(Revo_factor, "Revo_factor", id);

    db->ReadWords(rt_sync_the_speed, "rt_sync_the_speed", 0);

    db->LsRecords();
    db->Close();
    printf("[%d %d %d %d %d]\r\n", Vidle, Vpause, Vblked0, Vblked1, Vblked2);
    printf("[%d %d %d %d %d]\r\n", Vbr0, Vbr, Vrun, Vrunal, Rtime_base);
    printf("[%d %d %d %d %d]\r\n", Rtime_looser, Rtime_loose, Rtime_tight, Rtime_tightest, Rtime_spring);
    printf("[%d %d %d %d id=%d]\r\n", Revo_org_steps, Revo_startup_time, Revo_timespan, Revo_factor, id);
    printf("Loaded the parameters from the File. Vidle=%fV\r\n", Vidle*REF_VOLT_RATE);
    is_params_valid=1;
    mtr_works=1;
    current_id=id;
    return true;
}

void ListParams(cDataBase *db, String &str){
    int maxi;
    unsigned v, av;
    GET_VOLT(v); av=v; // 1
    db->Open();
    if ((maxi=db->FindMaxIdEx("Vidle"))==-1){db->Close(); return;}
    GET_VOLT(v); av+=v; // 2

    unsigned Revo_0, Revo_1, Revo_2, Revo_3, Revo_4;
    for (int i=0; i<=maxi; i++){
        if (!db->ReadWords(Revo_0, "Vidle", i)) continue;
        db->ReadWords(Revo_1, "Revo_org_steps", i);
        db->ReadWords(Revo_2, "Revo_startup_time", i);
        db->ReadWords(Revo_3, "Revo_timespan", i);
        db->ReadWords(Revo_4, "Revo_factor", i);
        str+=String("\r\nID: ")+i;
        str+=String("\r\nVidle: ")+Revo_0+"    "+(Revo_0*REF_VOLT_RATE)+"V";
        str+=String("\r\nRevo_org_steps: ")+Revo_1;
        str+=String("\r\nRevo_startup_time: ")+Revo_2;
        str+=String("\r\nRevo_timespan: ")+Revo_3;
        str+=String("\r\nRevo_factor: ")+Revo_4;
        str+=String("\r\n\r\n");
    }
    GET_VOLT(v); av=(v+av)/3; // 3
    str+=String("Current Voltage is: ")+av+"    "+(av*REF_VOLT_RATE)+"V";
    if (current_id>=0){
        str+=String("\r\nCurrent id is: ")+current_id;
    }
    else str+=String("\r\nCurrent parameters are not assigned id.");
    db->Close();
}


void loop_cmultipwms(){
    //static unsigned tmo1=micros();
    //if (cMultiPwms::Timeout(1500, tmo1)) cMultiPwms::traverseLinkedList();

    //PrintState();
    //LedBlinking();

    // 10 seconds after, load or probe the parameters.
    static unsigned tmo2=micros();
    if (cMultiPwms::Timeout(10000, tmo2, true)){
        cDataBase sdb(DBN);
        if (!LoadProbeParams(&sdb)) ProbeParams(MAX_TRIP_TIME);
        if (!is_params_valid) printf("motor is not supposed to work.\r\n");

        ManuallyControlMtr(2); // close first
    }

    // each 30 seconds, change params.
    static unsigned tmo5=micros();
    if (cMultiPwms::Timeout(30000, tmo5)){
        if (rt_sync_the_speed && mtr_works && is_params_valid){
            unsigned v;
            GET_VOLT(v);
            if (v<(current_voltage-Vbr0) || v>(current_voltage+Vbr0)){
                GET_VOLT(v);
                if (v<(current_voltage-Vbr0) || v>(current_voltage+Vbr0)){
                    cDataBase sdb(DBN);
                    if (LoadProbeParams(&sdb)) current_voltage=v;
                }
            }
        }
    }

    // do the probing if demand
    if (gen_params){
        gen_params=0;
        ProbeParams(MAX_TRIP_TIME);
        if (!is_params_valid){
            cDataBase sdb(DBN);
            if (!LoadProbeParams(&sdb)) printf("motor is not supposed to work.\r\n");
        }
    }

    // do the test if demand
    if (mtr_test){
        mtr_test=0;
        TestParams(MSTAGE, Revo_startup_time, Revo_timespan, Revo_factor);
    }

    // do the speed sync, after 30 seconds starts to work.
    static unsigned tmo3=micros();
    if (cMultiPwms::Timeout(30000, tmo3, true, true))
        if (rt_sync_the_speed && mtr_works && is_params_valid){
            MtrMovement(rt_the_speed/1000);
        }

    // after 5 minutes, disables mdns.
    static unsigned tmo4=micros();
    if (cMultiPwms::Timeout(300000, tmo4, true)) mdns_enabled=false;
}

void setup(){
    setup_cmultipwms();

    Serial.begin(115200);
    delay(5000);

    setup_IoT();

    Start();
    SetDC_all(99);

    GET_VOLT(current_voltage);
}

void loop(){
    static unsigned x=1;
    if (x && *remote_ap_ip){
        if (!*http_request_ip) strcpy(http_request_ip, remote_ap_ip);
        web_link=String("http://")+http_request_ip+"/s";
        x=0;
    }

    loop_IoT();
    loop_cmultipwms();
}
